/*
 *
 * Copyright (C) 2004 Alexander Yip (yipal@lcs.mit.edu)
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <sys/types.h>
#include "arpc.h"
#include "dbfe.h"
#include "bigint.h"
#include "str.h"
#include "parseopt.h"
#include "chord_types.h"
#include "rxx.h"
#include "misc_utils.h"
#include "id_utils.h"
#include <ihash.h>
#include "dhash_blockstore.h"
#include "dhash_types.h"
#include "audit_prot.h"
#include "merkle_misc.h"
#include "verify.h"

ptr<dhash_blockstore> storage;
str dbdir;

void
usage() 
{
  warn << "usage: aquadbtool -d db_directory [-p projkey(armor64A encoded)]\n";
  exit(1);
}

void printout( bigint *projkey) {
  bigint tempkey;
  vec<ptr<dbrec> > keylist;
  unsigned bytecount = 0;
  
    storage->getpkblockids(&keylist, projkey);
    for(unsigned i=0; i<keylist.size(); i++) {
      ptr<dbrec> data = storage->lookup_pkblock(keylist[i], &tempkey);
      long version = keyhash_version (data);
      bytecount += data->len;

      warn << "    " << armor64A(tempkey)
	   << "  " << armor64A(dbrec2id(keylist[i]))
	   << " 0x" << hexdump(keylist[i]->value, keylist[i]->len) 
	   << " ver " << version 
	   << " " << data->len << " bytes"
	   << "\n";
    }

    storage->getchblockids(&keylist, projkey);
    for(unsigned i=0; i<keylist.size(); i++) {
      ptr<dbrec> data = storage->lookup_chblock(keylist[i], &tempkey);
      bytecount += data->len;

      warn << "    " << armor64A(tempkey)
	   << "  " << armor64A(dbrec2id(keylist[i]))
	   << " 0x" << hexdump(keylist[i]->value, keylist[i]->len) 
	   << " " << data->len << " bytes"
	   << "\n";
    }
    warn << "total data contents: " << bytecount << "\n";
}

int
main (int argc, char **argv)
{
  bool projkeyinit = false;
  char ch;
  dbdir  = "";
  bigint projkey;

  while ((ch = getopt (argc, argv, "d:p:"))!=-1)
    switch (ch) {
    case 'd':
      dbdir = optarg;
      break;
    case 'p':
      projkey = dearmor64A_bigint(optarg);
      projkeyinit = true;
      break;
    default:
      usage ();
      break;
    }
  
  if (dbdir == "") {
    warn << "You must specify a database directory.\n";
    usage();
  }

  storage = New refcounted<dhash_blockstore>();
  storage->opendb(dbdir, false, 30);

  if (projkeyinit) {
    printout(&projkey);
  } else {
    printout(NULL);
  }

  return 0;
}
