/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef BRANCHTREE_H
#define BRANCHTREE_H

#include "stdinc.h"
#include "pack.h"
#include "record.h"

class pastdbfe;

struct btree_node {
  commit_info commit;
  commit_info prev;
  unsigned level;
  str proposed_name;
  vec<commit_info> next;
  void serialize (strbuf &b) const;
  static bool parse (xin &, btree_node &);
};

class btree_srv {
  /* btree_srv stores the branch tree as a collection of btree_nodes
   * in <_db>.  They link to each other via <prev> and <next> pointers.
   * to find a node, just look it up in <_db>.
   * 
   * btree_srv also stores other stuff like branch tags.
   */

private:
  ptr<pastdbfe> _db;
  void add_tag_aux (bigint c, str tag, bool tagged);
  void print_node(bigint n);

public:
  btree_srv (ptr<pastdbfe> db);
  void add_node (bigint c, btree_node &n);
  void add_tag (bigint c, str tag);
  str get_tags ();
  bool lookup_by_commit (bigint c, btree_node &n);
  bool lookup_by_tag (str t, bigint &c);
  bool find_join (bigint c0, bigint c1, bigint &cj);
  bool find_leaves (bigint c, vec<btree_node>& v);
  bool get_sole_descendant (bigint cr, bigint &head);

  void print(bigint start);
};

#endif

