/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef CFSR_PROTO_H
#define CFSR_PROTO_H

#include "cfs.h"

enum cfsr_proc {
  CFSRPROC_NULL     = 0,
  CFSRPROC_LOOKUP   = 1,
  CFSRPROC_GETATTR  = 2,
  CFSRPROC_FINDVER  = 3,
  CFSRPROC_READ     = 4,
  CFSRPROC_READDIR  = 5,
  CFSRPROC_READLINK = 6,
};

struct cfsr_attr {
  cfs_ftype type;
  uint64 size;
  uint32 mode;
  uint32 nlink;
  uint32 revision;
  bigint commit;
  bigint mode_commit;
};

struct cfsr_call_pair {
  void *arg;
  void *res;
};

struct cfsr_lookup_arg {
  bigint dir;
  str name;
};

struct cfsr_lookup_res {
  bigint object;
};

struct cfsr_getattr_arg {
  bigint object;
};

struct cfsr_getattr_res {
  ptr<cfsr_attr> attr;
};

struct cfsr_findver_arg {
  bigint object;
  unsigned revision;
};

struct cfsr_findver_res {
  bigint logid;
  bigint saddr;
  cfsr_findver_res () {}
  cfsr_findver_res (bigint l, bigint s)
    : logid (l), saddr (s) {}
  ~cfsr_findver_res () {}
};

struct cfsr_read_arg {
  bigint file;
  uint64 off;
  uint32 cnt;
};

struct cfsr_read_res {
  ptr<ns_block> data;
  bool eof;
};

struct cfsr_readdir_arg {
  bigint dir;
  uint64 cookie;
  uint32 count;
};

struct cfsr_readdir_res {
  ptr<cfs_dirlist> dir;
};

struct cfsr_readlink_arg {
  bigint file;
};

struct cfsr_readlink_res {
  str dest;
};

#endif

