#include "route_loctab.h"
#include "location.h"
#include "locationtable.h"
#include "misc_utils.h"
#include "loctabroute.h"
#include "blacklist.h"

route_loctab::route_loctab (ptr<vnode> vi, chordID xi, int _nsucc) :
  route_iterator(vi, xi), nsucc(_nsucc), r(NULL), filtering(false)
{
}


route_loctab::route_loctab (ptr<vnode> vi, chordID xi, rpc_program uc_prog,
			    int uc_procno, ptr<void> uc_args, int _nsucc) :
  route_iterator(vi, xi, uc_prog, uc_procno, uc_args), nsucc(_nsucc), 
  r(NULL), filtering(false)
{
}

route_loctab::route_loctab (ptr<vnode> vi, ptr<loctabroute> ri, chordID xi, 
			    chordID _projkey, int _nsucc) :
  route_iterator(vi, xi), 
  nsucc(_nsucc), r(ri), projkey(_projkey), filtering(true)
{
}


route_loctab::route_loctab (ptr<vnode> vi, ptr<loctabroute> ri, chordID xi, 
			    chordID _projkey, rpc_program uc_prog,
			    int uc_procno, ptr<void> uc_args, int _nsucc) :
  route_iterator(vi, xi, uc_prog, uc_procno, uc_args), 
  nsucc(_nsucc), r(ri), projkey(_projkey), filtering(true)
{
}

route_loctab::~route_loctab() 
{
  *deleted = true;
}

void
route_loctab::first_hop (cbhop_t cbi, ptr<chordID> guess)
{
  cb = cbi;
  next_hop();
}

void
route_loctab::next_hop () 
{
  ptr<location> loc, master, first, pred;
  //warn << "route_loctab::first_hop searching for " << x << "\n";
  successors_.clear();
  
  master = first = v->locations->closestsuccloc (x);
  while( !master->alive() || 
	 !is_willing (master->address(), projkey) ) {
    master = v->locations->next_loc_circular(master->id());

    if (master->id() == first->id()) {
      // no one is alive!
      warn << "no one is alive execpt me!\n";

      search_path.push_back(master);
      search_path.push_back(master);

      chord_node n;
      master->fill_node(n);
      successors_.push_back(n);

      //print_succs();
      (cb) (true);
      return; 
    }
  }

  // fill the search_path
  master = loc = v->locations->closestsuccloc (x);
  pred   = v->locations->prev_loc_circular (master->id());

  while(!pred->alive() && pred->id() != master->id() )
    pred   = v->locations->prev_loc_circular (pred->id());
  search_path.push_back(pred);
  search_path.push_back(master);

  int skipping = 0;
  while(loc && successors_.size() + skipping < (unsigned)nsucc ) {
    if (!loc->alive() && loc->down_duration() <= 60) {
      //warn << "Node " << loc->id() << " is temporarily down (" << loc->down_duration() << "). Skipping\n";
      skipping++;
    } else if (!loc->alive()) {
      //warn << "Node " << loc->id() << " is down (" << loc->down_duration() << "). Using next successor...\n";

    }
    if (loc->alive() && is_willing (loc->address(), projkey)) {
      chord_node n;
      loc->fill_node(n);
      successors_.push_back(n);
    }

    loc = v->locations->next_loc_circular ( loc->id() );
    if ( loc->id() == first->id() )
      break;
  }
  //print_succs();
  (cb) (true);
}

void
route_loctab:: send (ptr<chordID> guess) 
{
  warn << "route_loctab::send\n";
  assert(0); // Not implemented
}

ptr<location> 
route_loctab::pop_back () 
{
  return search_path.pop_back ();
}

void
route_loctab::print_succs() 
{
  warn << "successors for " << armor64A(x) << " " << x << "\n";
  if (filtering)
    warn << "   project " << armor64A(projkey) << " " << projkey << "\n";

  for(unsigned i=0; i<successors_.size(); i++) {
    warn << "  " << is_willing(successors_[i].r, projkey) 
	 << " " << successors_[i].x << "\n";
  }

  return;
}

bool
route_loctab::is_willing(const net_address &addr, bigint projkey) 
{
  if (!filtering)
    return true;
  else {
    return !blacklist::only().deny_project(addr, projkey);
  }
}
