/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef NS_H
#define NS_H

#include "async.h"
#include "record.h"
#include "sfscrypt.h"

class keyhash_payload;

struct ns_block {
  char *data;
  size_t len;
  ~ns_block () { delete [] data; }
  ns_block (const char *buf, size_t buflen)
    : data (New char[buflen]), len (buflen) {
    if (buf) memcpy (data, buf, len);
  }
};

enum ns_stat {
  NS_OK = 0,
  NS_ERR = 1,
  NS_NOENT = 2,
  NS_STALE = 3,
};

typedef callback<void, ns_stat, bigint>::ref cbinsert;
typedef callback<void, ptr<ns_block> >::ref cbblock;
typedef callback<void, ptr<keyhash_payload> >::ref cbkeyhash;
typedef callback<void, vec<bigint>, ns_stat>::ref cbcheckpks;
typedef callback<void, lptr, long, ns_stat>::ref cblptr;
typedef callback<void, crec, ns_stat>::ref cbcrec;

class ns : public virtual refcount {
public:
  static bigint pkid (ptr<sfspriv> sk);
  static ptr<ns> alloc (bigint);

  virtual ~ns () {}
  virtual void exit_cleanup () = 0;
  virtual void connect (cbv) = 0;
  virtual void connect (str, cbv) = 0;
  virtual bool connected () const = 0;
  virtual void sync () = 0;

  virtual bigint insert_local (const char *, size_t) = 0;
  virtual ptr<ns_block> retrieve_local (bigint n) = 0;
  virtual bool check_local (bigint n) = 0;

  virtual bigint insert_block
    (const char *, size_t, cbinsert cb, bool nocache = false) = 0;
  virtual void retrieve_block
    (bigint n, cbblock cb, bool nocache = false) = 0;

  virtual void insert_pk
    (ptr<sfspriv> sk, str data, cbinsert cb) = 0;
  virtual void insert_raw_pk_from_cache
    (bigint n, cbinsert cb) = 0;

  virtual void retrieve_pk (bigint n, cbkeyhash cb) = 0;
  virtual void retrieve_pk_from_net (bigint n, cbkeyhash cb) = 0;
  virtual void checkpks
    (vec<bigint> n, vec<long> v, cbcheckpks cb) = 0;

  // helpers
  virtual void retrieve_lptr (bigint n, cblptr cb) = 0;
  virtual void retrieve_commit (bigint n, cbcrec cb) = 0;

};

extern ptr<ns> nsobj;

#endif

