/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef PASTWATCH_DBFE_H
#define PASTWATCH_DBFE_H

#include "vec.h"
#include "ihash.h"
#include "bigint.h"
#include "dbfe.h"
#include <db.h>

/*
struct hashfn<DB *>
{
  hashfn () {}
  hash_t operator() (const DB *a) const { return (u_int)a; }
};

struct dbrec
{
  char *value;
  long len;

  dbrec (const void *v, long l) {
    len = l;
    value = New char[l];
    memcpy(value, v, l);
  }
  dbrec (const void *v, long l, unsigned s) {
    len = l;
    value = New char[l];
    memcpy (value+s, v, l-s);
  }
  ~dbrec() {  delete [] value; }
};


struct dbPair {
  ptr<dbrec> data;
  ptr<dbrec> key;
  
  dbPair(ref<dbrec> Key, ref<dbrec> Data) {
    data = Data;
    key = Key;
  };

  ~dbPair() {

  };

};

struct optionRec
{
  char *sig;
  long value;
};

struct dbOptions
{
  dbOptions();
  int addOption(char *optionSig, long value);
  long getOption(char *optionSig);

  vec<optionRec> options;
  int sel;
};
*/

struct pdbEnumeration
{
  pdbEnumeration(DB *db, DB_ENV *dbe);
  ~pdbEnumeration ();

  ptr<dbrec> getElement(u_int32_t flags, ptr<dbrec> startkey);
  ptr<dbrec> nextElement();
  ptr<dbrec> prevElement();
  ptr<dbrec> nextElement(ref<dbrec> startkey);
  ptr<dbrec> lastElement();
  ptr<dbrec> firstElement();

  DB* db_sync;
  DBC *cursor;
  char cursor_init;
};


class pastdbfe 
{
  DB_ENV* dbe;
  DB* db;
  bool _tx;
  int lockfd;
  str lockf;

 public:

  pastdbfe();
  ~pastdbfe();

  int opendb_env (const char *filename, dbOptions opts);
  int opendb (const char *filename, dbOptions opts);
  int opendb (dbOptions opts);
  int closedb ();
  
  int insert (ref<dbrec> key, ref<dbrec> data);
  ptr<dbrec> lookup (ref<dbrec> key);
  int remove (ref<dbrec> key);
  void sync ();
  ptr<pdbEnumeration> enumerate();
};

/*
inline ref<dbrec>
str2dbrec (str x)
{
  ptr<dbrec> q = New refcounted<dbrec> (x.cstr (), x.len ());
  return q;
}

inline str
dbrec2str (ptr<dbrec> r)
{
  str raw = str ((char *)r->value, r->len);
  return raw;
}
*/

inline ref<dbrec>
bigint2dbrec (bigint id)
{
  str whipme = id.getraw ();
  void *key = (void *)whipme.cstr ();
  int len = whipme.len ();

  ptr<dbrec> q = New refcounted<dbrec> (key, len);
  return q;
}

inline bigint
dbrec2bigint (ptr<dbrec> r)
{
  str raw = str ( (char *)r->value, r->len);
  bigint ret;
  ret.setraw (raw);
  return ret;
}

#endif

