/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef STDINC_H
#define STDINC_H

#include <time.h>
#include <sys/time.h>
#include "async.h"
#include "arpc.h"
#include "refcnt.h"
#include "xdrmisc.h"
#include "rpctypes.h"
#include "sfscrypt.h"
#include "bigint.h"
#include "sfsmisc.h"
#include "sha1.h"
#include "str.h"
#include "penviron.h"

template<>
struct hashfn<bigint> {
  hash_t operator() (const bigint &id) const { return id.getui (); }
};

inline str
armor64A (bigint n)
{
  char raw[20];
  mpz_get_rawmag_be (raw, sizeof (raw), &n);
  return armor64A (raw, sizeof (raw));
}

inline bool
dearmor64A_bigint (str x, bigint &n)
{
  str s = dearmor64A (x);
  if (!s)
    return false;
  mpz_set_rawmag_be (&n, s.cstr (), s.len ());
  return true;
}

inline bigint
compute_sha1_hash (const void *buf, size_t buflen)
{
  char h[sha1::hashsize];
  bzero(h, sha1::hashsize);
  sha1_hash (h, buf, buflen);

  bigint n;
  mpz_set_rawmag_be(&n, h, sha1::hashsize);  // For big endian
  return n;
}

inline char *
myctime (time_t tt)
{
  static char ts [32];
  struct tm *t = localtime (&tt);
  sprintf (ts, "%.2d/%.2d/%.2d %.2d:%.2d:%.2d",
           t->tm_mon+1,t->tm_mday,((t->tm_year+1900)%100),
	   t->tm_hour,t->tm_min,t->tm_sec);
  return ts;
}

inline u_int64_t
getusec ()
{
  timeval tv;
  gettimeofday (&tv, NULL);
  return tv.tv_sec * INT64(1000000) + tv.tv_usec;
}

inline str
pastwatch_tmpfile ()
{
  // create tmp file
  char raw[20];
  rnd.getbytes(raw, 20);
  str name = armor64A (raw, sizeof (raw));
  strbuf b;
  b << "/tmp/pastwatch." << name;
  str s = b;
  return s;
}

enum log_status {
  LOG_OK             = 0,
  LOGERR_ERR         = 1,
  LOGERR_IO          = 2,
  LOGERR_NOVIEW      = 100,
  LOGERR_NOCOMMIT    = 101,
  LOGERR_NOCPTR      = 102,
  LOGERR_NORECORD    = 103,
  LOGERR_CORRUPT     = 104,
  LOGERR_DIFFERENT   = 105,
  LOGERR_MISSINGDEP  = 106,
  LOGERR_EXCLCONT    = 107,
  LOGERR_ABORTED     = 108,
  LOGERR_TRYAGAIN    = 200,
};

typedef callback<void, log_status>::ptr cbl;

#endif

