/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef TRAVERSE_H
#define TRAVERSE_H

#include "stdinc.h"
#include "vec.h"
#include "ns.h"
#include "record.h"
#include "qhash.h"
#include "vervec.h"

class traverse;
class iterinfo;

typedef callback<void, iterinfo*>::ref cbtiter;
typedef callback<bool, bigint>::ref cbtc;

// a traverse object returns Commit records from the branch tree in
// partial order. the constructor takes a callback function. the
// callback is called for every commit record.  the traverse object
// passes a pointer to an iterinfo object as one of the arguments to
// the callback.  the iterinfo object is freed after the callback
// returns control to traverse.

class traverse : public virtual refcount { 
private:
  cbtiter _cb;
  cbtc _tc;
  comvec _start;

  unsigned _work;
  unsigned _lerr;
  vec<bigint> _todo;
  vec<crec> _commits;
  vec<bigint> _ids;
  qhash<bigint,unsigned> _bucket;

  void start ();
  void next ();
  void iter ();
  void load_commit (bigint x, crec r, ns_stat status);

public:
  traverse (cbtiter cb, const comvec& start, cbtc c);
  ~traverse () {}
};

class iterinfo {
public:
  log_status status;
  crec commit;
  bigint hash;
  bool finished;
  iterinfo (log_status s, crec r, bigint h, bool f)
    : status (s), commit (r), hash (h), finished (f)
  {}
};

#endif

