/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef VIEW_H
#define VIEW_H

#include "stdinc.h"
#include "pack.h"
#include "ns.h"
#include "vervec.h"

class view;

class view_constructor {
private:
  friend class view_impl;
  vec<bigint> _head;
  vec<str> _name;
  str _notify;
  void sort ();
  bool parse_old (xin *in);
  bool parse (xin *in);
  void add_entry_no_check (bigint u, str n);

public:
  view_constructor () { _notify = ""; }
  view_constructor (const view &);

  bool add_entry (bigint u, str n);
  void set_notify (str x);
  str serialize ();
  bigint store (ptr<sfspriv> sk, callback<void, ns_stat, bigint>::ref cb);
};

struct view_state {
  vervec pkv;
  comvec commits;
  vec<lptr> lptrs;
};

class view {
public:
  virtual unsigned nentries () const = 0;
  virtual bigint head (unsigned i) const = 0;
  virtual str name (unsigned i) const = 0;
  virtual str notify () const = 0;
  virtual void fetch (callback<void, ptr<view_state>, log_status>::ref cb) = 0;
  virtual void fetch (const vervec& pkv, const comvec& v,
                      callback<void, ptr<view_state>, log_status>::ref cb) = 0;
  static ptr<view> alloc (callback<void, log_status>::ref cb, bigint addr);
};

#endif

