/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include "pastwatch.h"
#include "opts.h"
#include "io.h"

extern ptr<cfsrw> _cfsrw;
extern str _myname;

void
rtag_cb (cbi cb, cfswcallp cc, cfsrw_status status)
{
  if (status)
    cb (-1);
  else
    cb (0);
}

void
rtag (str newtag, cbi cb)
{
  ptr<replica> r = _cfsrw->getreplica ();
      
  bigint n;
  if (arg_tag0 != "") {
    if (!dearmor64A_bigint (arg_tag0, n)) {
      warn << arg_tag0 << " is not a valid hash\n";
      cb (-1);
      return;
    }
    btree_node x;
    if (!r->btree ()->lookup_by_commit (n, x)) {
      warn << "cannot find " << arg_tag0 << " in branch tree\n";
      cb (-1);
      return;
    }
  }
  else
    n = 0;

  // see if tag exists already
  strbuf b;
  b << _myname << ":" << newtag;
  bigint c;
  if (r->btree ()->lookup_by_tag (b, c)) {
    warn << b << ": tag already exists\n";
    cb (-1);
    return;
  }
    
  cfswcallp cc = New refcounted<cfswcall> (CFSWPROC_TAG, wrap (rtag_cb, cb));
  cfsw_tag_arg *arg = static_cast<cfsw_tag_arg *> (cc->arg ());
  arg->commit = n;
  arg->tag = b;
  cfsw_call (cc);
}

