/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "cfs.h"
#include "cfsrw.h"  // cfsrw_write_file
#include "pastwatch.h"
#include "io.h"

class write_obj : public virtual refcount
{
  str path;
  bigint fh;
  callback<void, cfsrw_status>::ref cb;
 
  void
  write_done (cfswcallp c, cfsrw_status status)
  {
    if (status)
      cb (CFSRWERR_IO);
    else
      cb (CFSRW_OK);
  }

  void write_reply (cfsw_write_arg a, cfsrw_status status)
  {
    if (status) {
      cb (CFSRWERR_IO);
      return;
    }

    cfswcallp c = New refcounted<cfswcall>
      (CFSWPROC_WRITE, wrap (mkref(this), &write_obj::write_done));
    cfsw_write_arg *arg = static_cast<cfsw_write_arg *> (c->arg ());
    *arg = a;
    cfsw_call (c);
  }
 
public:
  write_obj (str path, bigint fh, bigint va,
             callback<void, cfsrw_status>::ref cb)
    : path (path), fh (fh), cb (cb)
  {
    cfsrw_write_file (path, fh, va, wrap (mkref(this), &write_obj::write_reply));
  }
};

extern mtime_t get_mtime (str);

mtime_t
write_file (str path, bigint fh, bigint va,
            callback<void, cfsrw_status>::ref cb)
{
  mtime_t m = get_mtime (path);
  New refcounted<write_obj>(path, fh, va, cb);
  return m;
}

